# opencv-python 라이브러리 설치
# 설치 명령어: pip install opencv-python
import cv2  # OpenCV: 이미지 처리 및 컴퓨터 비전 라이브러리

# numpy 라이브러리 설치
# 설치 명령어: pip install numpy
import numpy as np  # 수치 연산 라이브러리

# TensorFlow 및 Keras 라이브러리 설치
# 설치 명령어: pip install tensorflow
from keras.models import load_model  # Keras: 딥러닝 모델 로드

# Pillow(PIL) 라이브러리 설치
# 설치 명령어: pip install pillow
from PIL import Image, ImageOps  # Pillow: 이미지 처리 라이브러리

# NumPy 배열의 출력 형식을 설정
# 과학적 표기법(sci notation)을 비활성화하여 출력 값을 더 쉽게 확인
np.set_printoptions(suppress=True)

# 학습된 Keras 모델과 클래스 레이블 파일 로드
# "keras_model.h5" 파일과 "labels.txt" 파일은 사전에 준비되어 있어야 합니다.
model = load_model("Keras_model_motorcycle_stop.h5", compile=False)  # 학습된 Keras 모델 로드
class_names = open("labels1.txt", "r").readlines()  # 클래스 이름이 저장된 텍스트 파일 로드

# 이미지 전처리 함수 정의
def preprocess_frame(frame):
    """
    OpenCV 프레임을 PIL 이미지로 변환하고, Keras 모델 입력 형식에 맞게 전처리하는 함수.

    Parameters:
    frame (numpy.ndarray): OpenCV 프레임 (BGR 형식).

    Returns:
    numpy.ndarray: 모델 입력에 사용할 전처리된 NumPy 배열.
    """
    # OpenCV의 기본 컬러 형식인 BGR을 RGB로 변환
    image = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

    # NumPy 배열을 PIL 이미지로 변환
    image_pil = Image.fromarray(image)

    # PIL 이미지를 224x224 크기로 리사이징 (모델 입력 크기)
    # LANCZOS 알고리즘은 고품질 리사이징을 제공
    size = (224, 224)  # 리사이징할 크기 설정
    image_pil = ImageOps.fit(image_pil, size, Image.Resampling.LANCZOS)

    # PIL 이미지를 NumPy 배열로 변환
    image_array = np.asarray(image_pil)

    # 이미지 픽셀 값을 [0, 255] 범위에서 [-1, 1] 범위로 정규화
    normalized_image_array = (image_array.astype(np.float32) / 127.5) - 1

    # 모델 입력 형식에 맞게 배치 차원 추가 (1, 224, 224, 3)
    data = np.expand_dims(normalized_image_array, axis=0)

    return data

# 모델을 사용하여 프레임을 예측하는 함수 정의
def load_and_predict_frame(model, frame):
    """
    주어진 프레임에 대해 Keras 모델을 사용하여 예측하는 함수.

    Parameters:
    model: Keras 모델.
    frame (numpy.ndarray): OpenCV로 캡처한 프레임.

    Returns:
    str: 예측된 클래스 이름.
    float: 해당 클래스의 신뢰도 점수.
    """
    # 프레임 전처리 (PIL 변환, 리사이징, 정규화 등 수행)
    data = preprocess_frame(frame)

    # 모델을 사용하여 예측 수행
    prediction = model.predict(data)  # 예측 결과는 클래스별 확률 배열

    # 가장 높은 확률 값을 가진 클래스의 인덱스를 추출
    index = np.argmax(prediction)  # argmax는 가장 큰 값의 인덱스를 반환

    # 해당 클래스 이름과 신뢰도 점수를 추출
    class_name = class_names[index].strip()  # 클래스 이름 (공백 제거)
    confidence_score = prediction[0][index]  # 해당 클래스의 확률

    return class_name, confidence_score

# 예측할 이미지 파일 경로 설정
image_path = "st.png"  # 테스트할 이미지 파일의 경로

# OpenCV를 사용하여 이미지 파일 읽기
# imread 함수는 이미지를 BGR 형식으로 읽어들임
frame = cv2.imread(image_path)

# 이미지가 정상적으로 읽어졌는지 확인
if frame is None:  # 파일이 없거나 읽기에 실패한 경우
    print("이미지를 읽을 수 없습니다.")
else:
    # 이미지가 정상적으로 읽혔을 경우, OpenCV 창에 출력
    cv2.imshow('Image as Frame', frame)  # 창 제목: 'Image as Frame'
    cv2.waitKey(0)  # 키보드 입력 대기 (0이면 무한 대기)
    cv2.destroyAllWindows()  # 모든 OpenCV 창 닫기

    # 읽어들인 이미지를 학습된 모델로 예측
    class_name, confidence_score = load_and_predict_frame(model, frame)

    # 예측 결과 출력
    print(f"예측된 클래스: {class_name[2:]} (신뢰도: {confidence_score:.2f})")
  # 클래스 이름과 신뢰도 출력
