# bleak 라이브러리 설치
import asyncio
from bleak import BleakClient

# keyboard 라이브러리 설치
import keyboard  # 숫자키 입력을 감지하기 위한 라이브러리

# opencv-python 라이브러리 설치
import cv2

import numpy as np
# TensorFlow 라이브러리 설치
from keras.models import load_model
# Pillow 라이브러리 설치
from PIL import Image, ImageOps

# HM-10의 BLE 장치 MAC 주소
address = "18:93:D7:33:21:74"  # HM-10 모듈의 MAC 주소 입력

# HM-10 모듈에서 제어 명령을 전송할 특성(UUID)
CHARACTERISTIC_UUID = "0000ffe1-0000-1000-8000-00805f9b34fb"

# 자동차 제어 명령
commands = {
    '0': '0',
    '1': '1',
    '2': '2',
    '3': '3',
    '4': '4',
    '5': '5',
    '6': '6',
    '7': '7',
    '8': '8',
    '9': '9',
    '-': '*',
    '+': '#'
}

# DroidCam 스트림 URL 설정
ip_address = '192.168.0.9'  # DroidCam에서 확인한 IP 주소
port = '4747'  # DroidCam에서 사용하는 포트 (일반적으로 4747)
video_url = f'http://{ip_address}:{port}/video'

# OpenCV를 사용하여 스트림 열기
cap = cv2.VideoCapture(video_url)

if not cap.isOpened():
    print("스트림을 열 수 없습니다.")
    exit()

# NumPy 배열의 출력 옵션 과학적 표기법 비활성화
np.set_printoptions(suppress=True)

# 모델과 레이블 로드
model = load_model("keras_model.h5", compile=False)
class_names = open("labels.txt", "r").readlines()

def preprocess_frame(frame):
    """
    OpenCV 프레임을 PIL 이미지로 변환하고, Keras 모델 입력 형식에 맞게 전처리하는 함수.

    Parameters:
    frame (numpy.ndarray): OpenCV 프레임 (BGR 형식).

    Returns:
    numpy.ndarray: 모델 입력에 사용할 전처리된 NumPy 배열.
    """
    # BGR 형식에서 RGB 형식으로 변환
    image = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

    # NumPy 배열을 PIL 이미지로 변환
    image_pil = Image.fromarray(image)

    # 이미지를 224x224로 리사이징
    size = (224, 224)
    image_pil = ImageOps.fit(image_pil, size, Image.Resampling.LANCZOS)

    # 이미지를 NumPy 배열로 변환
    image_array = np.asarray(image_pil)

    # 이미지를 정규화: [0, 255] -> [-1, 1]
    normalized_image_array = (image_array.astype(np.float32) / 127.5) - 1

    # 배치 차원을 추가하여 모델 입력 형식에 맞춤
    data = np.expand_dims(normalized_image_array, axis=0)

    return data

def load_and_predict_frame(model, frame):
    """
    주어진 프레임에 대해 Keras 모델을 사용하여 예측하는 함수.

    Parameters:
    model: Keras 모델.
    frame (numpy.ndarray): OpenCV로 캡처한 프레임.

    Returns:
    str: 예측된 클래스 이름.
    float: 해당 클래스의 신뢰도 점수.
    """

    # 프레임 전처리
    data = preprocess_frame(frame)

    # 모델 예측
    prediction = model.predict(data)

    # 가장 높은 확률을 가진 클래스의 인덱스
    index = np.argmax(prediction)

    # 예측된 클래스 이름과 신뢰도 점수 반환
    class_name = class_names[index].strip()
    confidence_score = prediction[0][index]

    return class_name, confidence_score

# HM-10 BLE 모듈에 명령을 전송하는 함수
def send_command_sync(client, key):
    command = commands.get(key)
    if command:
        loop = asyncio.get_event_loop()
        loop.run_until_complete(client.write_gatt_char(CHARACTERISTIC_UUID, command.encode('utf-8')))
        print(f"'{command}' 명령 전송 완료")
    else:
        print("유효하지 않은 명령입니다.")

# 키 입력 감지 및 명령 전송 루프
def control_car_sync():
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    client = BleakClient(address)
    # loop.run_until_complete를 사용하여 BLE와의 비동기 통신을 동기적으로 처리
    loop.run_until_complete(client.connect())

    if client.is_connected:
        print(f"HM-10 ({address}) 연결 성공!")

        try:
            frame_counter = 0
            while True:
                # 숫자키 입력 감지 및 명령 전송
                for key in commands.keys():
                    if keyboard.is_pressed(key):
                        send_command_sync(client, key)

                # 종료를 위해 'q' 키를 누르면 제어 루프 종료
                if keyboard.is_pressed('q'):
                    print("프로그램을 종료합니다.")
                    break

                # 프레임 읽기
                ret, frame = cap.read()
                if ret:

                    # =====
                    # 원본 이미지의 가로, 세로 크기
                    (h, w) = frame.shape[:2]
                    # 90도 회전하는 회전 행렬 생성
                    center = (w // 2, h // 2)
                    rotation_matrix = cv2.getRotationMatrix2D(center, -90, 1.0)
                    # 회전 후 이미지 크기 계산
                    cos = np.abs(rotation_matrix[0, 0])
                    sin = np.abs(rotation_matrix[0, 1])
                    # 새로운 너비와 높이 계산
                    new_w = int((h * sin) + (w * cos))
                    new_h = int((h * cos) + (w * sin))
                    # 회전 행렬에 이동 보정 추가
                    rotation_matrix[0, 2] += (new_w / 2) - center[0]
                    rotation_matrix[1, 2] += (new_h / 2) - center[1]
                    # 가로, 세로를 바꾼 크기로 warpAffine 적용 (확보된 크기로 변환)
                    rotated_frame = cv2.warpAffine(frame, rotation_matrix, (new_w, new_h))
                    # 회전된 영상을 출력
                    cv2.imshow('Rotated Video', rotated_frame)
                    # =====

                    # 프레임을 화면에 표시
                    # cv2.imshow('DroidCam Video', frame)
                else:
                    print("프레임을 읽을 수 없습니다.")
                    break

                # 이벤트 루프를 실행하면서 특정 시간 동안 키보드 입력을 기다리는 역할('q' 키를 누르면 종료)
                if cv2.waitKey(1) & 0xFF == ord('q'):
                    break

                # 10 프레임마다 예측
                if frame_counter % 10 == 0:
                    class_name, confidence_score = load_and_predict_frame(model, frame)
                    label = f"{class_name}: {confidence_score:.2f}"
                    cv2.putText(frame, label, (10, 30), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0), 2)

                    print(class_name[2:], confidence_score)

                    if (class_name[2:] in ['Class 1']) and (confidence_score >= 0.98):
                        print('stop')
                        send_command_sync(client, '5')
                    else:
                        send_command_sync(client, '2')

                frame_counter += 1

            # 자원 해제
            cap.release()
            cv2.destroyAllWindows()

        finally:
            # 종료 시 BLE 연결 해제
            loop.run_until_complete(client.disconnect())
            loop.close()
    else:
        print(f"HM-10 ({address})에 연결할 수 없습니다.")

# 메인 실행 루프 (동기식으로 실행)
control_car_sync()